// Disk file (*.D64) lister for Commodore VC20/C64/C16/+4
// used to store data as in VC1541,1541 II,1551,1570,1571(single side)
// V1.02 (c) 2002 MicroWizard
// So you can use, copy, modify, forget it,
// as long, as you do not blame me. Have fun!
//
// Please try to compile it under a flat model operating environment
// I am lazy, so I read the whole D64 file into memory
// 
// Revision history:
// V0.00 2002.10.12 Start
// V1.00 2002.10.15 It simply works
// V1.01 2002.10.16 Handles '*' for whole directory listing
// V1.02 2002.10.17 Partially translates CBM ASCII codes to IBM PC ASCII codes

#include <stdio.h>
#include <dos.h>
#include <io.h>
#include <iostream.h>
#include <iomanip.h>

typedef unsigned char byte;
typedef unsigned short word;

byte cbmasciismallcaps[]         ="";
byte cbmasciicapsgrf[]  ="\x5b\x5c\x5d\x5e\x5f" "\x60\x61\x62" "\x71\x73\x78\x7a\x7b"
	"\xa0\xa2\xa6\xab\xad\xae" "\xb0\xb1\xb2\xb3\xbb\xbd" "";
byte ibmascii[]         ="\x5b\x9d\x5d\x18\x1b" "\xc4\x06\xb3" "\x07\x03\x05\x04\xc5"
	"\x5f\xdc\xb1\xc3\xc0\xbf" "\xda\xc1\xc2\xb4\xfe\xd9" "";

#pragma pack(push,1)
struct D64block
{
	byte	nexttrack;
	union
	{
		byte	nextsector;
		byte	lastbyte;	//if nexttrack equals to 0
	};
	byte	data[254];
};

struct D64file
{
	byte	type;
	byte	starttrack;
	byte	startsector;
	byte	name[16];
	byte	reltrack;
	byte	relsector;
	byte	reclen;
	byte	dummy[4];
	byte	newtrack;
	byte	newsector;
	word	blocknum;
};
#pragma pack(pop)

class D64
{
	static	byte	transtbl[256];
public:
	static	void	inittranslation(byte* charfrom,byte* charto);
	static	void	translate(char* dest, byte* src,int num);
private:
	class fdesc
	{
		byte	type;
		char	name[17];
		word	length;
	public:
		byte	gettype(){return type;};
		char*	getname(){return name;};
		word	getlength(){return length;};

		fdesc(){type=0;name[0]=0;length=0;};
		fdesc(D64file* fp)
		{
			type=fp->type;
			length=fp->blocknum;
			int i;
			for(i=0;i<16;i++)
			{
				if(fp->name[i]==0xa0)
				{
					break;
				}
			}
//			memcpy(name,fp->name,i);
			translate(name,fp->name,i);
			name[i]=0;
		};
	};

	FILE*	fhandle;
	byte*	disk;
	char	diskname[17];
	char	diskid[6];
	int		freeblocks;
	int		filenum;
	int		freefile;
	fdesc*	files;

	int			getdirectory(void);
	int			getfilenames(void);
	long		addressof(byte track,byte sector);
	int			addfiletolist(D64file* fp);

public:
				D64(char* filename);
				~D64();
	char*		name(void){return diskname;};
	char*		id(void){return diskid;};
	int			free(void){return freeblocks;};
	void		listfiles(ostream& s);
	void		dump(ostream& s);
	byte		filetype(word i);
	char*		filename(word i);
	word		filelen(word i);
	int			filenumber(void){return filenum;};
	D64block*	getblock(byte track,byte sector);
	D64block*	nextblock(D64block *bl);
};

byte D64::filetype(word i)
{
	if(i>=filenum)
	{
		return 0;
	}
	return files[i].gettype();
}

char* D64::filename(word i)
{
	if(i>=filenum)
	{
		return NULL;
	}
	return files[i].getname();
}

word D64::filelen(word i)
{
	if(i>=filenum)
	{
		return 0;
	}
	return files[i].getlength();
}

static byte D64::transtbl[256];
static void D64::inittranslation(byte* charfrom,byte* charto)
{
	byte i;
	for(i=0x0;i<0xff;i++)	//default initialization
	{
		transtbl[i]=i;
	}
	transtbl[0xff]=0xff;
	
	for(i=0x20;i<0x40;i++)		//standard signs
	{
		transtbl[i]=i;
	}
	for(i=0x40;i<0x5b;i++)		//letters
	{
		transtbl[i]=i;
	}
	for(i=0;charfrom[i];i++)	//special characters from parameters
	{
		transtbl[charfrom[i]]=charto[i];
	}

	for(i=0x0;i<0x20;i++)		//standard commodore assignment (silly???)
	{
		transtbl[0xc0+i]=transtbl[0x60+i];
		transtbl[0xe0+i]=transtbl[0xa0+i];
	}
}

void D64::translate(char* dest, byte* src,int num)
{
	for(int i=0;i<num;i++)
	{
		dest[i]=transtbl[src[i]];
	}
}

D64::D64(char* filename)
{
	disk=NULL;
	freeblocks=0;
	files=NULL;
	freefile=0;
	filenum=0;

	fhandle=fopen(filename,"rb");
	if(!fhandle)
	{
		return;
	}
	long len=filelength(fileno(fhandle));
	disk=new byte[len];
	if(!disk)
	{
		fclose(fhandle);
		fhandle=NULL;
		return;
	}
	long legyen=0L;
	long lett=0L;
	long pt=0L;
	do{
		legyen=(len>2000)?2000:len;
		lett=fread(disk+pt,1,legyen,fhandle);
		if(lett!=legyen)
		{
			delete disk;
			disk=NULL;
			fclose(fhandle);
			fhandle=NULL;
			return;
		}
		len-=lett;
		pt+=lett;
	}while(len>0L);
	fclose(fhandle);
	fhandle=NULL;
	if(getdirectory()>=0)
	{
		getfilenames();
	}
}

D64::~D64()
{
	if(fhandle)
	{
		fclose(fhandle);
		fhandle=NULL;
	}
	if(disk)
	{
		delete disk;
		disk=NULL;
	}
	if(files)
	{
		delete files;
		files=NULL;
	}
}

D64block* D64::getblock(byte track,byte sector)
{
	long p=addressof(track,sector);
	if(p>=0)
	{
		return((D64block*)(disk+p));
	}
	else
	{
		return NULL;
	}
}

D64block* D64::nextblock(D64block* bl)
{
	if(bl->nexttrack==0)
	{
		return NULL;
	}
	return(getblock(bl->nexttrack,bl->nextsector));
}

long D64::addressof(byte track,byte sector)
{
	if(track<1)
	{
		return -1;
	}
	else if(track<18)
	{
		if(sector<21)
		{
			return((0+21*(track-1)+sector)*256);
		}
		else
			return -1;
	}
	else if(track<25)
	{
		if(sector<19)
		{
			return((17*21+19*(track-18)+sector)*256);
		}
		else
			return -1;
	}
	else if(track<31)
	{
		if(sector<18)
		{
			return((17*21+7*19+18*(track-25)+sector)*256);
		}
		else
			return -1;
	}
	else if(track<35)
	{
		if(sector<17)
		{
			return((17*21+7*19+6*18+17*(track-31)+sector)*256);
		}
		else
			return -1;
	}
	else
		return -1;
}

int D64::getdirectory(void)
{
	D64block* blk=getblock(18,0);	// BAM

	if(blk==NULL)
	{
		return -1;
	}

	int i;
	freeblocks=0;
	for(i=0;i<35;i++)
	{
		if(i!=18-1)	// track 18 does not count
		{
			freeblocks+=blk->data[2+i*4];
		}
	}

//	memcpy(diskname,blk->data+142,16);
	translate(diskname,blk->data+142,16);
	diskname[16]=0;

//	memcpy(diskid,blk->data+160,5);
	translate(diskid,blk->data+160,5);
	diskid[5]=0;

	return freeblocks;
}

int D64::getfilenames(void)
{
	D64block* blk=NULL;

	byte trk=18;	// start of directory
	byte sec=1;

	do
	{
		blk=getblock(trk,sec);	// start of directory
		if(blk==NULL)
		{
			return -1;
		}
		int i;
		D64file* fp=NULL;
		for(i=0;i<8;i++)
		{
			fp=(D64file*)(blk->data+i*32);
			if(fp->type==0)
			{
				break;
			}
			if(freefile==0)
			{
				fdesc* tmp=new fdesc[filenum+50];
				int j;
				for(j=0;j<filenum;j++)
				{
					tmp[j]=files[j];
				}
				delete files;
				files=tmp;
				freefile+=50;
			}
			files[filenum++]=fdesc(fp);
			freefile--;
		}
		if(fp->type==0)
		{
			break;
		}
		trk=blk->nexttrack;
		sec=blk->nextsector;
	}while(trk!=0);

	return 0;
}

void D64::listfiles(ostream& s)
{
	char f[50];
	f[0]='"';
	char t[6];
	t[5]=0;
	char l[6];
	l[5]=0;
	for(int i=0;i<filenumber();i++)
	{
		f[1]=0;
		strcat(f,filename(i));
		strcat(f,"\"                ");
		f[18]=0;
		byte q=filetype(i);
		t[0]=(q&0x80)?' ':'*';
		switch(q&0x07)
		{
			case 0:
				t[1]='D';
				t[2]='E';
				t[3]='L';
				break;
			case 1:
				t[1]='S';
				t[2]='E';
				t[3]='Q';
				break;
			case 2:
				t[1]='P';
				t[2]='R';
				t[3]='G';
				break;
			case 3:
				t[1]='U';
				t[2]='S';
				t[3]='R';
				break;
			case 4:
				t[1]='R';
				t[2]='E';
				t[3]='L';
				break;
			default:		// VC1541 creates a special kind of garbage, which I don't want to emulate
				t[1]='?';
				t[2]='?';
				t[3]='?';
		}
		t[4]=(q&0x40)?'<':' ';
		sprintf(l,"%-5d",filelen(i));
		s << l << f << t << endl;
	}
}

void D64::dump(ostream& s)
{
	s << endl << "0 " << '"' << name() << '"' << " " << id() << endl;
	listfiles(s);
	s << free() << " BLOCKS FREE." << endl;
}
/****************************************************************************************/

void main(int argc, char** argv)
{
	cout << "D64 file lister V1.02" << endl;
	cout << "(C) 2002 MicroWizard, Bitraspers, Tom from HardworX" << endl;

	if(argc==1)
	{
		cout << endl << "Usage: d64dir <file1.d64> [<file2.d64> [...]]" << endl;
		cout << "or" << endl;
		cout << "       d64dir *" << endl;
		return;	
	}

	D64::inittranslation(cbmasciicapsgrf,ibmascii);

	D64* disk=NULL;
	if(argc==2 && argv[1][0]=='*' && argv[1][1]==0)
	{
		//directory listing
		cout << "Starting directory scan..." << endl;
		char* path="*.d64";
		struct FIND* f=findfirst(path,_A_RDONLY | _A_ARCH);
		int i=0;
		while(f!=NULL)
		{
			i++;
			cout << endl << "Found:" << f->name << endl;
			disk=new D64(f->name);
			disk->dump(cout);
			delete disk;
			f=findnext();
		}
		cout << endl << "Total " << i << " D64 file" <<  ((i>1)?"s ":" ") << "listed." << endl;
	}
	else
	{
		for(int i=1;i<argc;i++)
		{
			disk=new D64(argv[i]);
			disk->dump(cout);
			delete disk;
		}
	}
}